# coding=utf-8
"""
"""
from __future__ import absolute_import

import os
import re
import logging
import tempfile
import itertools
import fnmatch
from functools import partial
from cStringIO import StringIO
from datetime import datetime
from urllib import quote
from zipfile import ZipFile, is_zipfile

from werkzeug.exceptions import InternalServerError
from xlwt import Workbook, easyxf
import whoosh.query as wq
from flask import (g, redirect, request, make_response, flash,
                   current_app, send_file, jsonify,
                   render_template_string, Markup, render_template)

import sqlalchemy as sa
from sqlalchemy import func

from abilian.i18n import _, _n
from abilian.core.models.subjects import User, Group
from abilian.core.extensions import db
from abilian.core.signals import activity
from abilian.services.security import security, Role, READ, WRITE
from abilian.web.action import actions
from abilian.web.views import default_view

from abilian.web import csrf, http, url_for

from abilian.sbe.apps.communities.views import default_view_kw
from ..repository import repository
from ..models import Folder, Document, icon_for, icon_url
from ..search import reindex_tree
from .util import (
  get_folder, check_manage_access, get_document,
  check_read_access, breadcrumbs_for, check_write_access, create_document,
  get_selected_objects, edit_object, get_new_filename,
)
from .views import documents


route = documents.route


@route("/")
def index():
  folder = g.community.folder
  url = url_for(folder)
  return redirect(url)


@default_view(documents, Folder, id_attr='folder_id', kw_func=default_view_kw)
@route("/folder/<int:folder_id>")
def folder_view(folder_id):
  folder = get_folder(folder_id)
  bc = breadcrumbs_for(folder)
  actions.context['object'] = folder
  ctx = dict(folder=folder,
             children=folder.filtered_children,
             breadcrumbs=bc,
             csrf_token=csrf.field(), )
  return render_template("documents/folder.html", **ctx)


@route("/folder/<int:folder_id>/json")
def folder_json(folder_id):
  """ return parent folder + subfolders
  """
  folder = get_folder(folder_id)
  folder_url = partial(url_for, '.folder_json')
  result = {}
  has_permission = security.has_permission
  result['current_folder_selectable'] = has_permission(g.user, WRITE, folder,
                                                       inherit=True)
  folders = result['folders'] = []
  bc = result['breadcrumbs'] = []
  subfolders = sorted((f for f in folder.subfolders
                       if has_permission(g.user, READ, f, inherit=True)),
                      key=lambda f: f.title)

  parent = folder

  # breadcrumbs
  for parent in reversed(list(folder._iter_to_root())):
    if parent.is_root_folder:
      continue

    data = {
      'id': parent.id,
      'url': folder_url(folder_id=parent.id,
                        community_id=parent.community.slug),
      'title': parent.title,
    }
    bc.append(data)

  if folder.parent is not None and not folder.parent.is_root_folder:
    # not at root folder: allow to go 1 level up
    data = bc[-2].copy()
    data['title'] = ".."
    folders.append(data)
  else:
    result['current_folder_selectable'] = False

  for folder in subfolders:
    data = {
      'id': folder.id,
      'url': folder_url(folder_id=folder.id,
                        community_id=folder.community.slug),
      'title': folder.title,
    }
    folders.append(data)

  return jsonify(result)


@route("/folder/<int:folder_id>/members")
def members(folder_id):
  folder = get_folder(folder_id)
  bc = breadcrumbs_for(folder)
  actions.context['object'] = folder
  members = folder.members()
  ctx = dict(folder=folder, members=members, breadcrumbs=bc)
  return render_template("community/members.html", **ctx)


@route("/folder/<int:folder_id>/permissions")
@http.nocache
def permissions(folder_id):
  folder = get_folder(folder_id)
  check_manage_access(folder)

  bc = breadcrumbs_for(folder)
  actions.context['object'] = folder
  local_roles_assignments = folder.get_local_roles_assignments()
  principals = set((p for p, r in local_roles_assignments))
  security._fill_role_cache_batch(principals)

  users_and_local_roles = [(user, role, repository.has_access(user, folder))
                           for user, role in local_roles_assignments
                           if isinstance(user, User)]
  groups_and_local_roles = [t for t in local_roles_assignments
                            if isinstance(t[0], Group)]

  users_and_inherited_roles = groups_and_inherited_roles = ()

  if folder.inherit_security:
    inherited_roles_assignments = folder.get_inherited_roles_assignments()
    users_and_inherited_roles = [(user, role, False)
                                 for user, role in inherited_roles_assignments
                                 if isinstance(user, User)]
    groups_and_inherited_roles = [t for t in inherited_roles_assignments
                                  if isinstance(t[0], Group)]

  query = Group.query
  query = query.order_by(func.lower(Group.name))
  all_groups = query.all()

  class EntryPresenter(object):
    _USER_FMT = (u'<a href="{{ url_for("social.user", user_id=user.id) }}">'
                 '{{ user.name }}</a>')
    _GROUP_FMT = (u'<a href="{{ url_for("social.group_home", group_id=group.id)'
                  ' }}">{{ group.name }}</a>')

    def __init__(self, e):
      render = render_template_string
      self.entry = e
      self.date = e.happened_at.strftime('%Y-%m-%d %H:%M')
      self.manager = render(
        u'<img src="{{ user_photo_url(e.manager, size=16) }}" alt="" />'
        '<a href="{{ url_for("social.user", user_id=e.manager_id) }}">'
        '{{ e.manager.name }}</a>', e=e)

      if e.op == e.SET_INHERIT:
        msg = _(u'On {date}, {manager} has activated inheritance')
      elif e.op == e.UNSET_INHERIT:
        msg = _(u'On {date}, {manager} has deactivated inheritance')
      elif e.op == e.GRANT:
        msg = _(u'On {date}, {manager} has given role "{role}" to {principal}')
      elif e.op == e.REVOKE:
        msg = _(u'On {date}, {manager} has revoked role "{role}" from '
                '{principal}')
      else:
        raise Exception("Unknown audit entry type %s" % e.op)

      principal = u''
      if self.entry.user:
        principal = render(self._USER_FMT, user=self.entry.user)
      elif self.entry.group:
        principal = render(self._GROUP_FMT, group=self.entry.group)

      self.msg = Markup(msg.format(date=self.date, manager=self.manager,
                                   role=self.entry.role,
                                   principal=principal))

  audit_entries = [EntryPresenter(e) for e in security.entries_for(folder)]
  csrf_token = csrf.field()

  ctx = dict(folder=folder,
             users_and_local_roles=users_and_local_roles,
             users_and_inherited_roles=users_and_inherited_roles,
             groups_and_local_roles=groups_and_local_roles,
             groups_and_inherited_roles=groups_and_inherited_roles,
             audit_entries=audit_entries,
             csrf_token=csrf_token,
             all_groups=all_groups,
             breadcrumbs=bc)

  return render_template("documents/permissions.html", **ctx)


@route("/folder/<int:folder_id>/permissions", methods=['POST'])
@csrf.protect
def permissions_update(folder_id):
  folder = repository.get_folder_by_id(folder_id)
  check_manage_access(folder)
  has_permission = security.has_permission
  action = request.form.get("action")

  if action in ("activate_inheritance", "deactivate_inheritance"):
    inherit_security = (action == "activate_inheritance")

    if not (inherit_security or
            has_permission(g.user, 'manage', folder, inherit=False)):
      # don't let user shoot himself in the foot
      flash(_(u'You must have the "manager" local role on this folder in '
              'order to deactivate inheritance.'), u'error')
      return redirect(url_for(".permissions",
                              folder_id=folder_id,
                              community_id=folder.community.slug))

    security.set_inherit_security(folder, inherit_security)
    db.session.add(folder)
    reindex_tree(folder)
    db.session.commit()
    return redirect(url_for(".permissions",
                            folder_id=folder_id,
                            community_id=folder.community.slug))

  elif action == "add-user-role":
    role = request.form.get("role").lower()
    user_id = int(request.form.get("user"))
    user = User.query.get(user_id)

    security.grant_role(user, role, folder)
    reindex_tree(folder)
    db.session.commit()
    return redirect(url_for(".permissions",
                            folder_id=folder_id,
                            community_id=folder.community.slug))

  elif action == "add-group-role":
    role = request.form.get("role").lower()
    group_id = int(request.form.get("group"))
    group = Group.query.get(group_id)

    security.grant_role(group, role, folder)
    reindex_tree(folder)
    db.session.commit()
    return redirect(url_for(".permissions",
                            folder_id=folder_id,
                            community_id=folder.community.slug))

  else:
    action, args = request.form.items()[0]
    role, object_id = args.split(":")
    role = role.lower()
    object_id = int(object_id)

    if action == 'delete-user-role':
      user = User.query.get(object_id)
      # remove role in a subtransaction, to prevent manager shoot himself in the
      # foot
      transaction = db.session.begin_nested()
      security.ungrant_role(user, role, folder)

      if (user == g.user and role == 'manager'
          and not has_permission(g.user, 'manage', folder, inherit=True)):

        transaction.rollback()
        flash(_(u'Cannot remove "manager" local role for yourself: you '
                'don\'t have "manager" role (either by security inheritance '
                'or by group membership)'),
              'error')
      else:
        reindex_tree(folder)
        transaction.commit()
        flash(_(u"Role {role} for user {user} removed on folder {folder}"
        ).format(role=role, user=user.name, folder=folder.name),
              "success")
    elif action == 'delete-group-role':
      group = Group.query.get(object_id)
      # remove role in a subtransaction, to prevent manager shoot himself in the
      # foot
      transaction = db.session.begin_nested()
      security.ungrant_role(group, role, folder)

      if (role == 'manager'
          and not has_permission(g.user, 'manage', folder, inherit=True)):
        transaction.rollback()
        flash(_(u'Cannot remove "manager" local role for group "{group}": you'
                ' don\'t have "manager" role by security inheritance or by '
                'local role').format(group=group.name),
              'error')
      else:
        flash(_(u"Role {role} for group {group} removed on folder {folder}"
        ).format(role=role, group=group.name, folder=folder.name),
              "success")
        reindex_tree(folder)
        transaction.commit()

    db.session.commit()
    return redirect(url_for(".permissions",
                            folder_id=folder_id,
                            community_id=folder.community.slug))


@route("/folder/<int:folder_id>/permissions_export")
@http.nocache
def permissions_export(folder_id):
  folder = repository.get_folder_by_id(folder_id)
  check_manage_access(folder)

  wb = Workbook()
  ws = wb.add_sheet("Sheet 1")

  cols = [
    (u'Accès', 20),
    (u'Identifiant', 40),
    (u'Prénom', 14),
    (u'Nom', 20),
    (u'Rôle', None),
    (u'Local', None),
    (u'Héritage', None),
    (u'Communauté', 60),
  ]

  # styling
  # from xlwt doc: width unit is 1/256 of '0' from first font in excel file
  for c, width in enumerate((c[1] for c in cols)):
    if width is not None:
      ws.col(c).width = 256 * width

  ws.row(0).height = 256 + 128

  ws.panes_frozen = True
  ws.remove_splits = True
  ws.horz_split_pos = 1

  header_style = easyxf(
    'font: bold true;'
    'alignment: horizontal center, vertical center;'
  )
  for c, val in enumerate((c[0] for c in cols)):
    ws.write(0, c, val, header_style)

  # data
  permissions = iter_permissions(folder, g.user)
  row_offset = 0
  current_community = None
  for r, row in enumerate(permissions, 1):
    if current_community is None:
      current_community = row[-1]

    if current_community != row[-1]:
      current_community = row[-1]
      row_offset += 1

    # data grouping enter (refered as 'outline' in xlwt documention)
    ws.row(r + row_offset).level = 1

    for c, value in enumerate(row):
      if isinstance(value, Role):
        value = unicode(value)
      ws.write(r + row_offset, c, value)

    # data grouping exit
    ws.row(r + row_offset).level = 1

  debug = request.args.get('debug_sql')
  if debug:
    # useful only in DEBUG mode, to get the debug toolbar in browser
    return '<html><body>Exported</body></html>'

  fd = StringIO()
  wb.save(fd)

  response = make_response(fd.getvalue())
  response.headers['content-type'] = 'application/ms-excel'
  folder_name = folder.title.replace(u' ', u'_')
  file_date = datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
  filename = u'permissions-{}-{}.xls'.format(folder_name, file_date)
  content_disposition = u'attachment;filename="{}"'.format(filename)
  response.headers['content-disposition'] = content_disposition
  return response


def iter_permissions(folder, user):
  """ Iterator returning permissions settings on folder and its subfolders tree.
  """
  if not security.has_permission(user, "manage", folder, inherit=True):
    raise StopIteration

  community = folder.path
  local_roles = frozenset(folder.get_local_roles_assignments())
  inherited_roles = frozenset((folder.get_inherited_roles_assignments()
                               if folder.inherit_security
                               else []))

  result = {}
  for principal, role in (local_roles | inherited_roles):
    data = result.setdefault((principal, role), {})
    data['local'] = (principal, role) in local_roles
    data['inherit'] = (principal, role) in inherited_roles

  def _sort_key(item):
    """ Sorts by name, groups first. """
    principal = item[0][0]
    is_user = isinstance(principal, User)
    item_key = [is_user]
    if is_user:
      last_name = principal.last_name if principal.last_name else u''
      first_name = principal.first_name if principal.first_name else u''
      item_key.append(last_name.lower())
      item_key.append(first_name.lower())
    else:
      item_key.append(principal.name)
    return item_key

  for (p, role), data in sorted(result.items(), key=_sort_key):
    is_user = isinstance(p, User)
    has_access = False if is_user else '*'
    identifier = p.email if is_user else u'* Group *'
    first_name = p.first_name if is_user else u'-'
    last_name = p.last_name if is_user else p.name
    local = data['local']
    inherit = data['inherit']

    yield (has_access, identifier, first_name, last_name, role,
           local, inherit, community)

  subfolders = (f for f in folder.subfolders
                if security.has_permission(user, "manage", folder))

  for subfolder in subfolders:
    for permission in iter_permissions(subfolder, user):
      yield permission


#
# Actions on folders
#
@route("/folder/<int:folder_id>", methods=['POST'])
@csrf.protect
def folder_post(folder_id):
  """
  A POST on a folder can result on several different actions (depending on the
  `action` parameter).
  """
  folder = get_folder(folder_id)
  action = request.form.get("action")

  if action == 'edit':
    return folder_edit(folder)

  elif action == 'upload':
    return upload_new(folder)

  elif action == 'download':
    return download_multiple(folder)

  elif action == 'delete':
    return delete_multiple(folder)

  elif action == 'new':
    return create_subfolder(folder)

  elif action == 'move':
    return move_multiple(folder)

  elif action == 'change-owner':
    return change_owner(folder)

  else:
    # Probably an error or a hack attempt.
    # Logger will inform sentry if enabled
    logger = logging.getLogger(__name__)
    logger.error(u"Unknown folder action.", extra={'stack': True})
    flash(_(u"Unknown action."), "error")
    return redirect(url_for(folder))


def folder_edit(folder):
  check_write_access(folder)

  changed = edit_object(folder)

  if changed:
    db.session.commit()
    flash(_(u"Folder properties successfully edited."), "success")
  else:
    flash(_(u"You didn't change any property."), "success")
  return redirect(url_for(folder))


ARCHIVE_IGNORE_FILES = {
  u'__MACOSX/*',
  u'.DS_Store',
}
# translates patterns to match with any parent directory ((*/)?pattern should
# match)
ARCHIVE_IGNORE_FILES = {re.compile(u'(?:.*\\/)?' + fnmatch.translate(pattern))
                        for pattern in ARCHIVE_IGNORE_FILES}

# skip directory names. Directory will be created only if they contains files
ARCHIVE_IGNORE_FILES.add(re.compile(fnmatch.translate(u'*/')))


def explore_archive(fd, filename=None, uncompress=False):
  """
  Given an uploaded file descriptor, return it or a list of archive
  content.

  Yield tuple(filepath, file-like object), where filepath is a list
  whose components are directories and last one is filename.
  """
  if filename is None:
    filename = fd.filename

  if not isinstance(filename, unicode):
    filename = unicode(fd.filename, errors='ignore')

  if not uncompress:
    yield [], fd
    raise StopIteration

  if is_zipfile(fd):
    with ZipFile(fd, 'r') as archive:
      for zipinfo in archive.infolist():
        filename = zipinfo.filename
        if isinstance(filename, bytes):
          # not unicode: try to convert from utf-8 (OSX case: unicode flag not
          # set), then legacy cp437
          # http://stackoverflow.com/questions/13261347/correctly-decoding-zip-entry-file-names-cp437-utf-8-or
          try:
            filename = filename.decode('utf-8')
          except UnicodeDecodeError:
            filename = filename.decode('cp437')

        if any(pattern.match(filename) is not None
               for pattern in ARCHIVE_IGNORE_FILES):
            continue

        filepath = filename.split(u'/')
        filename = filepath.pop()
        zip_fd = archive.open(zipinfo, 'r')
        setattr(zip_fd, 'filename', filename)
        setattr(zip_fd, 'content_type', None)
        yield filepath, zip_fd
  else:
    yield [], fd

def upload_new(folder):
  check_write_access(folder)
  session = current_app.db.session()
  base_folder = folder
  uncompress_files = 'uncompress_files' in request.form
  fds = request.files.getlist('file')
  created_count = 0
  path_cache = {} # mapping folder path in zip: folder instance

  for upload_fd in fds:
    for filepath, fd in explore_archive(upload_fd, uncompress=uncompress_files):
      folder = base_folder
      parts = []
      # traverse to final directory, create intermediate if necessary. Folders
      # may be renamed if a file already exists, path_cache is used to keep
      # track of this
      for subfolder_name in filepath:
        parts.append(subfolder_name)
        path = u'/'.join(parts)
        if path in path_cache:
          folder = path_cache[path]
          continue

        subfolders = {f.title: f for f in folder.subfolders}
        if subfolder_name in subfolders:
          folder = subfolders[subfolder_name]
          path_cache[path] = folder
          continue

        subfolder_name = get_new_filename(folder, subfolder_name)
        folder = folder.create_subfolder(subfolder_name)
        session.flush()
        path_cache[path] = folder

      create_document(folder, fd)
      created_count += 1

  flash(_n(u"One new document successfully uploaded",
           u"%(num)d new document successfully uploaded",
           num=created_count),
        "success")

  session.commit()
  return redirect(url_for(folder))


def download_multiple(folder):
  folders, docs = get_selected_objects(folder)

  def rel_path(path, content):
    return u'{}/{}'.format(path, content.title)

  def zip_folder(zipfile, folder, path=u''):
    for doc in folder.documents:
      doc_path = rel_path(path, doc)
      zipfile.writestr(doc_path, doc.content)

    for subfolder in folder.filtered_subfolders:
      zip_folder(zipfile, subfolder, rel_path(path, subfolder))
    return zipfile

  # if using upstream send file: just create a temps file.
  # if app is streaming itself: use NamedTemporaryFile so that file is removed
  # on close()
  temp_factory = (tempfile.mktemp if current_app.use_x_sendfile
                  else tempfile.NamedTemporaryFile)
  zip_fn = temp_factory(prefix='tmp-' + current_app.name + '-',
                        suffix='.zip')
  with ZipFile(zip_fn, "w") as zipfile:
    for doc in docs:
      zipfile.writestr(doc.title, doc.content)
    for subfolder in folders:
      zip_folder(zipfile, subfolder, subfolder.title)

  if not isinstance(zip_fn, str):
    zip_fn.seek(0, os.SEEK_END)
    size = zip_fn.tell()
    zip_fn.seek(0)
  else:
    size = os.path.getsize(zip_fn)

  resp = send_file(zip_fn, mimetype="application/zip", as_attachment=True,
                   attachment_filename=quote(
                     folder.title.encode("utf8") + ".zip"))
  resp.headers.add('Content-Length', str(size))
  return resp


def delete_multiple(folder):
  check_write_access(folder)

  folders, docs = get_selected_objects(folder)

  for obj in docs + folders:
    app = current_app._get_current_object()
    community = g.community._model
    activity.send(app, actor=g.user, verb="delete", object=obj, target=community)
    repository.delete_object(obj)

  if docs + folders:
    db.session.commit()
    if docs and folders:
      msg = _(u"%(file_num)d files and %(folder_num)d folders sucessfully "
              "deleted.", file_num=len(docs), folder_num=len(folders))
    elif docs and not folders:
      msg = _n("1 file sucessfully deleted.",
               "%(num)d files sucessfully deleted.",
               num=len(docs))
    else:
      msg = _n("1 folder sucessfully deleted.",
               "%(num)d folders sucessfully deleted.",
               num=len(folders))

    flash(msg, "success")
  else:
    flash(_(u"No object deleted"), "error")

  return redirect(url_for(folder))


def move_multiple(folder):
  folders, docs = get_selected_objects(folder)
  count_f = len(folders)
  count_d = len(docs)
  current_folder_url = url_for(folder)

  if not (count_f + count_d):
    flash(_(u'Move elements: no elements selected.'), 'info')
    return redirect(current_folder_url)

  try:
    target_folder_id = int(request.form.get('target-folder'))
  except ValueError:
    flash(_(u'Move elements: no destination folder selected. Aborted.'),
          'error')
    return redirect(current_folder_url)

  target_folder = repository.get_folder_by_id(target_folder_id)

  if folder == target_folder:
    flash(_(u'Move elements: source and destination folder are identical,'
            ' nothing done.'), 'error')
    return redirect(current_folder_url)

  if not security.has_permission(g.user, 'write', folder, inherit=True):
    # this should not happen: this is just defensive programming
    flash(_(u'You are not allowed to move elements from this folder'), 'error')
    return redirect(current_folder_url)

  if not security.has_permission(g.user, 'write', target_folder, inherit=True):
    flash(_(u'You are not allowed to write in folder "{folder}"'
    ).format(folder=target_folder.title),
          'error')
    return redirect(current_folder_url)

  for item in itertools.chain(folders, docs):
    # FIXME: maybe too brutal
    check_write_access(item)

  # verify we are not trying to move a folder inside itself or one of its
  # descendants
  f = target_folder
  while f:
    if f in folders:
      flash(_(u'Move elements: destination folder is included in moved '
              'elements. Moved nothing.'), 'error')
      return redirect(url_for(folder))
    f = f.parent

  exist_in_dest = []
  for item in itertools.chain(folders, docs):
    try:
      with db.session.begin_nested():
        item.parent = target_folder
    except sa.exc.IntegrityError:
      exist_in_dest.append(item)

  if exist_in_dest:
    # items existing in destination: cancel operation
    db.session.rollback()
    msg = _(u'Move elements: canceled, some elements exists in destination '
            u'folder: {elements}')
    elements = u', '.join(u'"{}"'.format(i.title) for i in exist_in_dest)
    flash(msg.format(elements=elements), 'error')
    return redirect(current_folder_url)

  db.session.commit()

  msg_f = (_n(u'1 folder', u'{count} folders', count_f) if count_f
           else  _(u'0 folder')).format(count=count_f)

  msg_d = (_n(u'1 document', u'{count} documents', count_d)
           if count_d
           else _(u'0 document')).format(count=count_d)

  msg = _(u'{folders} and {documents} moved to {target}').format(
    folders=msg_f, documents=msg_d, target=target_folder.title)
  flash(msg, 'success')

  return redirect(url_for(folder))


def create_subfolder(folder):
  check_write_access(folder)

  title = request.form.get("title", u"")
  description = request.form.get("description", u"")
  subfolder = folder.create_subfolder(title)
  subfolder.description = description

  db.session.commit()
  return redirect(url_for(folder))


def change_owner(folder):
  check_manage_access(folder)
  items = itertools.chain(*get_selected_objects(folder))

  user_id = request.form.get('new_owner', type=int)
  user = User.query.get(user_id)

  for item in items:
    item.owner = user

  db.session.commit()
  return redirect(url_for(folder))


@route("/folder/check_valid_name")
@http.nocache
def check_valid_name():
  """ Check if name is valid for content creation in this folder
  """
  object_id = int(request.args.get('object_id'))
  action = request.args.get('action')
  title = request.args.get('title')

  get_object = get_document if action == 'document-edit' else get_folder
  obj = get_object(object_id)
  check_read_access(obj)

  if action == u'new':
    parent = obj
    help_text = _(u'An element named "{name}" is already present in folder')
  elif action in (u'folder-edit', u'document-edit'):
    parent = obj.parent
    help_text = _(u'Cannot rename: "{name}" is already present in parent '
                  'folder')
  else:
    raise InternalServerError()

  existing = set((e.title for e in parent.children))

  if action in (u'folder-edit', u'document-edit'):
    try:
      existing.remove(obj.title)
    except KeyError:
      pass

  result = {}
  valid = result['valid'] = title not in existing
  if not valid:
    result['help_text'] = help_text.format(name=title)

  return jsonify(result)


@route("/folder/<int:folder_id>/descendants")
def descendants_view(folder_id):
  folder = get_folder(folder_id)
  bc = breadcrumbs_for(folder)
  actions.context['object'] = folder

  root_path_ids = folder._indexable_parent_ids + u'/{}'.format(folder.id)
  svc = current_app.services['indexing']
  filters = wq.And([wq.Term('community_id', folder.community.id),
                    wq.Term('parent_ids', root_path_ids),
                    wq.Or([wq.Term('object_type', Folder.entity_type),
                           wq.Term('object_type', Document.entity_type),]),
                  ])

  results = svc.search(u'', filter=filters, limit=None)
  by_path = {}
  owner_ids = set()
  for hit in results:
    by_path.setdefault(hit['parent_ids'], []).append(hit)
    owner_type, owner_id = hit['owner'].split(u':')
    if owner_type == u'user':
      try:
        owner_id = int(owner_id)
        owner_ids.add(owner_id)
      except ValueError:
        pass

  for children in by_path.values():
    children.sort(key=lambda hit: (hit['object_type'] != Folder.entity_type,
                                   hit['name'].lower()))

  descendants = []
  def visit(path_id, level=0):
    children = by_path.get(path_id, ())

    for child in children:
      is_folder = child['object_type'] == Folder.entity_type
      type_letter = u'F' if is_folder else u'D'
      descendants.append((level, type_letter, child))

      if is_folder:
        path_id = child['parent_ids'] + u'/{}'.format(child['id'])
        visit(path_id, level+1)

  visit(root_path_ids, 0)

  owners = dict()
  owners_q = User.query\
                 .filter(User.id.in_(owner_ids)) \
                 .add_column(sa.sql.func.concat('user:', User.id).label('key'))
  for user, key in owners_q:
    owners[key] = user

  ctx = dict(folder=folder,
             descendants=descendants,
             owners=owners,
             breadcrumbs=bc,
             get_icon=get_icon_for_hit,
             csrf_token=csrf.field(), )
  return render_template("documents/descendants.html", **ctx)


def get_icon_for_hit(hit):
  if hit['object_type'] == Folder.entity_type:
    return icon_url('folder.png')

  content_type = hit['content_type']
  icon = icon_for(content_type)
  return icon



# coding=utf-8
from os.path import join, dirname

from flask import (render_template, redirect, flash, url_for, request,
                   g, make_response, jsonify, current_app)
from flask_babel import gettext as _
from werkzeug.exceptions import NotFound, InternalServerError
from sqlalchemy import func
from abilian.core.models.subjects import User, Group
from abilian.core.extensions import db
from abilian.web import csrf
from abilian.web.views import default_view
from abilian.services.image import resize, CROP
from abilian.services.security import security

from ..forms import GroupForm
from .social import social
from .util import Env


DEFAULT_GROUP_MUGSHOT = open(join(dirname(__file__),
                                 "../../../static/images/frog.jpg")).read()


@social.route("/groups/")
def groups():
  tab = request.args.get("tab", "all_groups")
  e = Env()
  if tab == 'all_groups':
    e.groups = Group.query.order_by(Group.name).all()
    if not security.has_role(g.user, "admin"):
      e.groups = [ group for group in e.groups
                   if group.public or g.user in group.members ]
  else:
    e.groups = g.user.groups
    e.groups.sort(lambda x, y: cmp(x.name, y.name))
  return render_template("social/groups.html", **e)


def is_admin(group):
  is_admin = g.user in group.admins
  if not is_admin and 'security' in current_app.extensions:
    is_admin = current_app.services['security'].has_role(g.user, 'admin')

  return is_admin


@social.route("/groups/<int:group_id>")
@default_view(social, Group, 'group_id')
def group_home(group_id):
  e = Env(csrf_token=csrf.field())
  e.group = Group.query.get(group_id)
  e.is_admin = is_admin(e.group)
  return render_template("social/group.html", **e)


@social.route("/groups/<int:group_id>/json")
def group_json(group_id):
  members = Group.query.get(group_id).members
  q = request.args.get("q", u'').lower()
  if q:
    members = filter(
      lambda u: any([term.startswith(q)
                    for name in (u.first_name.lower(), u.last_name.lower())
                    for term in name.split()]),
      members)

  result = {'results': [{'id': obj.id, 'text': obj.name}
                        for obj in members]}
  return jsonify(result)


@social.route("/groups/<int:group_id>", methods=['POST'])
@csrf.protect
def group_post(group_id):
  group = Group.query.get(group_id)
  action = request.form.get('action')
  return_url = request.form.get('return_url')

  membership_actions = frozenset(
    ('add', 'remove',
     'add-admin', 'remove-admin',
     ))

  if action not in ('join', 'leave'):
    assert is_admin(group)

  if action == 'join':
    g.user.join(group)
  elif action == 'leave':
    g.user.leave(group)
  elif action in membership_actions:
    user_id = request.form.get('user', u'').strip()
    try:
      user_id = int(user_id)
    except:
      flash(_(u'Error: No user selected'), 'error')
      return redirect(url_for(".group_home", group_id=group_id))

    user = User.query.get(user_id)

    if action == 'add':
      user.join(group)
    elif action == 'remove':
      user.leave(group)
    elif action == 'add-admin':
      if user not in group.admins:
        group.admins.append(user)
    elif action == 'remove-admin':
      if user in group.admins:
        group.admins.remove(user)
  else:
    raise Exception("Should not happen")
  db.session.commit()

  if return_url:
    # TODO: security check
    return redirect(return_url)
  else:
    return redirect(url_for(".group_home", group_id=group_id))


@social.route("/groups/new")
def groups_new():
  # TODO later
  return

  e = Env()
  e.form = GroupForm()
  return render_template("social/groups-new.html", **e)


@social.route("/groups/new", methods=['POST'])
def groups_new_post():
  # TODO later
  return

  form = GroupForm()

  if form.validate():
    group = Group()
    form.populate_obj(group)
    db.session.add(group)
    db.session.commit()
    flash(_(u"Your new group has been created"), category='info')
    return redirect(url_for('.group_home', group_id=group.id))

  else:
    e = Env()
    e.form = form
    return render_template("social/groups-new.html", **e)


# TODO: duplicated code (with user_mugshot). Extract common method.
@social.route("/groups/<int:group_id>/mugshot")
def group_mugshot(group_id):
  size = int(request.args.get('s', 55))
  if size > 500:
    raise ValueError("Error, size = %d" % size)
  group = Group.query.get(group_id)

  if not group:
    raise NotFound()

  if group.photo:
    data = group.photo
  else:
    data = DEFAULT_GROUP_MUGSHOT

  if size:
    data = resize(data, size, size, mode=CROP)

  response = make_response(data)
  response.headers['content-type'] = 'image/jpeg'
  response.headers.add('Cache-Control', 'public, max-age=600')
  return response


#
# Ad-hoc JSON endpoints, used by select boxes
#
@social.route("/groups/json")
def groups_json():
  q = request.args.get("q").replace("%", " ").lower()

  if not q or len(q) < 2:
    raise InternalServerError()

  query = Group.query
  #query = query.filter(func.lower(Group.name).like(q + "%"))
  query = query.order_by(func.lower(Group.name))
  all = query.all()

  result = {'results': [{'id': obj.id, 'text': obj.name} for obj in all]}
  return jsonify(result)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import sqlalchemy as sa

from abilian.services.security import security, READ

from .models import CmisObject, Folder, Document


class SecurityException(Exception):
  pass


class Repository(object):
  """
  A simple document repository, implementing the basic functionalities of the
  CMIS model.
  """

  def __init__(self, app=None):
    if app is not None:
      self.init_app(app)

  def init_app(self, app):
    self.app = app
    app.extensions['content_repository'] = self

  @property
  def root_folder(self):
    folder =  Folder.query.filter(Folder.parent == None).first()
    if folder:
      return folder

    # Should only happen during tests
    folder = Folder(title=u'root')
    return folder

  def get_object(self, id=None, path=None):
    """
    Gets the CMIS object (document or folder) with either the given `id` or the
    given `path`. Returns None if the object doesn't exist.
    """
    if id:
      return self.get_object_by_id(id)
    else:
      return self.get_object_by_path(path)

  #
  # Id based navigation
  #
  def get_object_by_id(self, id):
    """
    Gets the CMIS object (document or folder) with the given `id`.

    Returns None if the object doesn't exist.
    """
    obj = CmisObject.query.get(id)
    if obj is not None and not isinstance(obj, CmisObject):
      return None
    return obj

  def get_folder_by_id(self, id):
    """
    Gets the folder with the given `id`.

    Returns None if the folder doesn't exist.
    """
    return Folder.query.get(id)

  def get_document_by_id(self, id):
    """
    Gets the document with the given `id`.

    Returns None if the document doesn't exist.
    """
    document = Document.query.get(id)
    return document

  #
  # Path based navigation
  #
  def get_object_by_path(self, path):
    """
    Gets the CMIS object (document or folder) with the given `path`.

    Returns None if the object doesn't exist.
    """
    return self.root_folder.get_object_by_path(path)

  def get_folder_by_path(self, path):
    """
    Gets the folder with the given `path`.

    Returns None if the folder doesn't exist.
    """
    obj = self.root_folder.get_object_by_path(path)
    if obj is None or not obj.is_folder:
      return None
    else:
      return obj

  def get_document_by_path(self, path):
    """
    Gets the document with the given `path`.

    Returns None if the document doesn't exist.
    """
    obj = self.root_folder.get_object_by_path(path)
    if obj is None or not obj.is_document:
      return None
    else:
      return obj

  #
  # COPY / MOVE support
  #
  def copy_object(self, obj, dest_folder, dest_title=None):
    new_obj = obj.clone(title=dest_title, parent=dest_folder)
    if obj.is_folder:
      for child in obj.children:
        self.copy_object(child, new_obj)
    return new_obj

  def move_object(self, obj, dest_folder, dest_title=None):
    obj.parent = dest_folder
    if dest_title:
      obj.title = dest_title

  def rename_object(self, obj, title):
    obj.title = title

  def delete_object(self, obj):
    if obj.is_root_folder:
      raise Exception("Can't delete root folder.")

    session  = sa.orm.object_session(obj)
    setattr(obj, '__path_before_delete', obj.path) # for audit log.
    parent = obj.parent
    collection = parent.subfolders if obj.is_folder else parent.documents
    session.delete(obj)
    collection.remove(obj)

  #
  # Locking (TODO)
  #
  def is_locked(self, obj):
    return False

  def can_unlock(self, obj):
    return True

  def lock(self, obj):
    return "???"

  def unlock(self, obj):
    pass

  #
  # Security / access rights
  #
  def has_permission(self, user, permission, obj):
    return security.has_permission(user, permission, obj, inherit=True)

  def has_access(self, user, obj):
    """
    Checks that user has actual right to reach this object, 'read' permission
    on each of object's parents.
    """
    current = obj
    while current.parent is not None:
      if not self.has_permission(user, READ, current):
        return False
      current = current.parent
    return True


repository = Repository()



# coding=utf-8

from unittest import TestCase

from abilian.core.entities import all_entity_classes
from abilian.core.models.subjects import User, Group

from ..models import Message, PrivateMessage


def check_editable(object):
  if hasattr(object, '__editable__'):
    for k in object.__editable__:
      assert hasattr(object, k)


class TestUsers(TestCase):

  def test_user(self):
    user = User(first_name=u"John", last_name=u"Test User",
                email=u"test@example.com", password="toto")
    check_editable(user)

    assert u"John Test User" == user.name
    assert u"John Test User" == unicode(user)
    #self.assertEquals(len(user.messages), 0)

  def test_user_follow(self):
    # id is provided by DB (which is not used in this test), and is required for
    # having in (user1 != user2) == True
    user1 = User(id=1, first_name=u"John", last_name=u"Test User 1",
                 email=u"test1@example.com", password="toto")
    user2 = User(id=2, first_name=u"Joe", last_name=u"Test User 2",
                 email=u"test2@example.com", password="toto")

    assert len(user1.followers) == 0
    assert len(user1.followees) == 0
    assert len(user2.followers) == 0
    assert len(user2.followees) == 0

    user1.follow(user2)

    assert user1.is_following(user2)
    assert len(user2.followers) == 1
    assert len(user1.followees) == 1
    assert len(user2.followees) == 0
    assert len(user1.followers) == 0
    assert user2 in user1.followees

    user1.unfollow(user2)

    assert not user1.is_following(user2)
    assert len(user1.followers) == 0
    assert len(user1.followees) == 0
    assert len(user2.followers) == 0
    assert len(user2.followees) == 0
    assert user2 not in user1.followers

  def test_group(self):
    user = User(first_name=u"John", last_name=u"Test User",
                email=u"test@example.com", password="toto")
    group = Group(name=u"Group 1")

    user.join(group)

    assert user.is_member_of(group)
    assert len(group.members) == 1
    assert len(user.groups) == 1
    assert group.members == {user}
    assert user.groups == {group}

    user.leave(group)

    assert not user.is_member_of(group)
    assert len(group.members) == 0
    assert len(user.groups) == 0


class TestContent(TestCase):

  def test_private_message(self):
    pm = PrivateMessage(creator_id=0, recipient_id=0)
    check_editable(pm)

  def test_status_update(self):
    user = User(first_name=u"John", last_name=u"Test User",
                email=u"test@example.com", password="toto")
    #self.assertEquals(len(user.messages), 0)

    message = Message()
    message.author = user
    check_editable(message)

    #self.assertEquals(len(user.messages), 1)
    #self.assertEquals(user.messages[0], message)
    #self.assertEquals(message.author_id, user.uid)

  def test_tags(self):
    m = Message(content=u"abc #123 #cde #voilà_l_été #789")
    assert m.tags == [u'123', u'cde', u'voilà_l_été', u'789']

  def test_get_all_entity_classes(self):
    classes = all_entity_classes()
    assert Message in classes
    assert PrivateMessage in classes



# coding=utf-8
"""
"""
from __future__ import absolute_import

import difflib
from os.path import dirname, join
from urllib import quote

from flask import g, render_template, redirect, request, \
  flash, current_app, make_response
from markdown import markdown
from markupsafe import Markup
import sqlalchemy as sa
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.exceptions import NotFound
from whoosh.searching import Hit
from abilian.i18n import _, _n, _l
from abilian.core.signals import activity
from abilian.core.extensions import db
from abilian.web.action import actions
from abilian.web.views import default_view, ObjectView, ObjectEdit, ObjectCreate
from abilian.web.util import url_for
from abilian.web import csrf, nav

from abilian.sbe.apps.communities.blueprint import Blueprint
from abilian.sbe.apps.communities.views import (
  default_view_kw as community_dv_kw,
)
from .forms import WikiPageForm
from .models import WikiPage, WikiPageAttachment, WikiPageRevision

wiki = Blueprint("wiki", __name__,
                 url_prefix="/wiki",
                 template_folder="templates")
route = wiki.route


@wiki.url_value_preprocessor
def init_wiki_values(endpoint, values):
  g.current_tab = 'wiki'

  endpoint = nav.Endpoint('wiki.index', community_id=g.community.slug)
  g.breadcrumb.append(nav.BreadcrumbItem(label=_l(u'Wiki'), url=endpoint))

  title = request.args.get('title', u'').strip()
  if title and title != 'Home':
    g.breadcrumb.append(
      nav.BreadcrumbItem(label=title,
                         url=nav.Endpoint('wiki.page',
                                          community_id=g.community.slug,
                                          title=title)))


@route('/')
def index():
  return redirect(url_for(".page", title='Home', community_id=g.community.slug))


def wiki_page_default_view_kw(kw, obj, obj_type, obj_id, **kwargs):
  kw = community_dv_kw(kw, obj, obj_type, obj_id, **kwargs)
  title = kwargs.get('title')
  if title is None:
    if isinstance(obj, (Hit, dict)):
      title = obj.get('name')
    else:
      title = obj.title
    kw['title'] = title
  return kw


#
# Page level (all functions prefixed by 'page_').
#
class BasePageView(object):
  Model = WikiPage
  pk = 'title'
  Form = WikiPageForm
  base_template = 'wiki/_base.html'

  @property
  def is_home_page(self):
    return self.obj is not None and self.obj.title == u'Home'

  @property
  def template_kwargs(self):
    """
    Template render arguments. You can override `base_template` for
    instance. Only `view` and `form` cannot be overriden.
    """
    kw = super(BasePageView, self).template_kwargs
    kw['page'] = self.obj
    return kw

  def init_object(self, args, kwargs):
    title = kwargs['title'] = request.args['title'].strip()
    if title:
      try:
        self.obj = get_page_by_title(title)
      except NoResultFound:
        if title == u'Home':
          self.obj = create_home_page()
        else:
          flash(_(u"This page doesn't exit. You must create it first."),
                 "warning")
          self.redirect(
            url_for(".page_new", title=title, community_id=g.community.slug)
          )
      actions.context['object'] = self.obj
    return args, kwargs

  def index_url(self):
    return url_for(".index", community_id=g.community.slug)

  def view_url(self):
    return url_for(self.view_endpoint,
                   community_id=g.community.slug,
                   title=self.obj.title)


class PageView(BasePageView, ObjectView):
  """
  """
  template = 'wiki/page.html'
  view_endpoint = '.page'
  decorators = [
    default_view(wiki, WikiPage,
                 id_attr=None,
                 kw_func=wiki_page_default_view_kw)
  ]

  def init_object(self, args, kwargs):
    args, kwargs = BasePageView.init_object(self, args, kwargs)
    if not self.obj:
      title = kwargs['title']
      if title == u'Home':
        self.obj = create_home_page()
      else:
        return redirect(
          url_for(".page_edit", title=title, community_id=g.community.slug))

    actions.context['object'] = self.obj
    return args, kwargs

route('/page/')(PageView.as_view('page'))


class PageEdit(BasePageView, ObjectEdit):
  # template = 'wiki/page_edit.html'
  title = _("Edit page")
  last_revision = None
  _message_success = _l(u"Wiki page successfully edited.")

  def init_object(self, args, kwargs):
    if request.method != 'POST':
      return super(PageEdit, self).init_object(args, kwargs)

    page_id = request.form.get('page_id') or None
    if page_id is not None:
      page_id = int(page_id)
      self.obj = WikiPage.query.get_or_404(page_id)
      actions.context['object'] = self.obj

    return args, kwargs

  def get_form_kwargs(self):
    kwargs = ObjectEdit.get_form_kwargs(self)
    kwargs['page_id'] = kwargs['last_revision_id'] = None
    if self.obj is not None and self.obj.id:
      # if no 'id', then it's a new object (PageCreate). We shouldn't call
      # obj.last_revision, since it will issue a flush (thus creating obj and
      # the 1st revision, resulting in an "edit conflict")
      kwargs['page_id'] = self.obj.id
      kwargs['last_revision_id'] = self.obj.last_revision.id
    return kwargs

  def prepare_args(self, args, kwargs):
    super(PageEdit, self).prepare_args(args, kwargs)
    last_revision_id = self.form.last_revision_id.data
    if last_revision_id:
      self.last_revision = WikiPageRevision.query\
          .filter(WikiPageRevision.page == self.obj,
                  WikiPageRevision.id == last_revision_id)\
          .first()
    return args, kwargs

  def before_populate_obj(self):
    self.redirect_if_no_change()
    form = self.form
    self.obj.create_revision(form.body_src.data, form.message.data)

  def redirect_if_no_change(self):
    form = self.form
    if all(f.data == f.object_data for f in (form.title, form.body_src)):
      flash(_(u"You didn't make any change to this page."))
      return self.redirect(url_for(self.obj))

  @property
  def activity_target(self):
    return self.obj.community

  def form_invalid(self):
    is_conflict = self.form.errors.pop('last_revision_id', None)
    if not self.form.errors and is_conflict is not None:
      # no other error on form, just edit conflict: show new text
      self.form.last_revision_id.errors = []
      field = self.form.body_src
      current = self.obj.last_revision
      self.form.last_revision_id.data = current.id  # update edited revision
      self.redirect_if_no_change()  # same edition? don't bother

      if (self.last_revision.body_src == current.body_src
          and self.form.validate()):
        # only title change? cannot show diff: save if valid
        return self.form_valid()
      else:
        edited_src = field.data
        field.data = current.body_src
        edited_diff = [
          l for l in difflib.ndiff(self.last_revision.body_src.splitlines(True),
                                  edited_src.splitlines(True))
          if not l[0] == u'?']
        current_diff = [
          l for l in difflib.ndiff(self.last_revision.body_src.splitlines(True),
                                  current.body_src.splitlines(True))
          if not l[0] == u'?']
        field.errors.append(
            Markup(render_template('wiki/edit_conflict_error.html',
                                   current=current,
                                   current_diff=current_diff,
                                   edited_diff=edited_diff)))

    return None


route('/edit')(PageEdit.as_view('page_edit', view_endpoint='.page'))


class PageCreate(PageEdit, ObjectCreate):
  title = _l("Create page")
  _message_success = _l(u"Wiki page successfully created.")

  get_form_kwargs = ObjectCreate.get_form_kwargs

  def init_object(self, args, kwargs):
    args, kwargs = ObjectCreate.init_object(self, args, kwargs)
    self.obj.community = g.community
    session = sa.orm.object_session(self.obj)
    if session:
      sa.orm.session.make_transient(self.obj)
      for rev in self.obj.revisions:
        sa.orm.session.make_transient(rev)
    return args, kwargs


route('/new')(PageCreate.as_view('page_new', view_endpoint='.page'))


@route('/source/')
def page_source():
  title = request.args['title'].strip()
  try:
    page = get_page_by_title(title)
  except NoResultFound:
    return redirect(
      url_for(".page_edit", title=title, community_id=g.community.slug))

  actions.context['object'] = page
  return render_template('wiki/source.html', page=page)


@route('/changes/')
def page_changes():
  title = request.args['title'].strip()
  try:
    page = get_page_by_title(title)
  except NoResultFound:
    return redirect(
      url_for(".page_edit", title=title, community_id=g.community.slug))
  revisions = page.revisions
  revisions = sorted(revisions, key=lambda x: -x.number)
  actions.context['object'] = page
  return render_template('wiki/changes.html', page=page, revisions=revisions)


@route('/compare/')
def page_compare():
  title = request.args['title'].strip()
  try:
    page = get_page_by_title(title)
  except NoResultFound:
    return redirect(
      url_for(".page_edit", title=title, community_id=g.community.slug))
  revisions = page.revisions
  revisions = sorted(revisions, key=lambda x: x.number)
  revs_to_compare = []
  for arg in request.args:
    if arg.startswith("rev"):
      revs_to_compare.append(int(arg[3:]))
  if len(revs_to_compare) != 2:
    flash(_(u"You must check exactly 2 revisions."), "error")
    return redirect(url_for(".page_changes", title=title, community_id=g.community.slug))

  revs_to_compare.sort()
  from_rev = revisions[revs_to_compare[0]]
  to_rev = revisions[revs_to_compare[1]]
  assert from_rev.number == revs_to_compare[0]
  assert to_rev.number == revs_to_compare[1]

  from_lines = from_rev.body_src.splitlines(1)
  to_lines = to_rev.body_src.splitlines(1)

  differ = difflib.Differ(charjunk=difflib.IS_CHARACTER_JUNK)
  diff = differ.compare(from_lines, to_lines)
  diff = [line for line in diff if not line.startswith("?")]

  actions.context['object'] = page
  return render_template('wiki/compare.html', page=page, diff=diff,
                         rev1=from_rev, rev2=to_rev,)


@route('/delete/', methods=["POST"])
def page_delete():
  title = request.form['title'].strip()
  try:
    page = get_page_by_title(title)
  except NoResultFound:
    flash(_(u"This page doesn't exist"), "error")
    return redirect(url_for(".index", community_id=g.community.slug))

  db.session.delete(page)

  app = current_app._get_current_object()
  community = g.community._model
  activity.send(app, actor=g.user, verb="delete", object=page, target=community)

  db.session.commit()
  flash(_(u"Page %(title)s deleted.", title=title))
  return redirect(url_for(".index", community_id=g.community.slug))


@route('/attachments')
def attachment_download():
  title = request.args['title'].strip()
  attachment_id = int(request.args['attachment'])
  try:
    page = get_page_by_title(title)
  except NoResultFound:
    raise NotFound()

  attachment = WikiPageAttachment.query.get(attachment_id)
  assert attachment is not None and attachment.wikipage is page

  response = make_response(attachment.content)
  response.headers['content-length'] = attachment.content_length
  response.headers['content-type'] = attachment.content_type
  content_disposition = (
    'attachment;filename="{}"'.format(quote(attachment.name.encode('utf8')))
                        )
  response.headers['content-disposition'] = content_disposition
  return response


@route('/attachments', methods=['POST'])
@csrf.protect
def attachment_upload():
  title = request.args['title'].strip()
  try:
    page = get_page_by_title(title)
  except NoResultFound:
    raise NotFound()

  files = request.files.getlist('attachments')
  saved_count = 0

  if len(files) > 0:
    for f in files:
      name = f.filename
      if not isinstance(name, unicode):
        name = unicode(f.filename, encoding='utf-8', errors='ignore')

      if not name:
        continue

      attachment = WikiPageAttachment(name=name)
      attachment.wikipage = page
      attachment.set_content(f.read(), f.content_type)
      db.session.add(attachment)
      saved_count += 1

    db.session.commit()

  if saved_count:
    flash(_n(u"One new document successfully uploaded",
             u"%(num)d new document successfully uploaded",
             count=saved_count, num=len(files)),
          "success")
  else:
    flash(_(u'No file uploaded.'))

  return redirect(url_for(page))


@route('/attachments/delete', methods=['POST'])
@csrf.protect
def attachment_delete():
  title = request.args['title'].strip()
  attachment_id = int(request.args['attachment'])
  try:
    page = get_page_by_title(title)
  except NoResultFound:
    raise NotFound()

  attachment = WikiPageAttachment.query.get(attachment_id)
  assert attachment is not None and attachment.wikipage is page

  if request.form.get('action') == 'delete':
    name = attachment.name
    db.session.delete(attachment)
    db.session.commit()
    flash(_(u'Attachment "{name}" has been deleted').format(name=name))

  return redirect(url_for(page))


#
# Wiki-level (prefixed by 'wiki_')
#
@route('/pages')
def wiki_pages():
  pages = WikiPage.query \
    .filter(WikiPage.community_id == g.community.id) \
    .order_by(WikiPage.title)\
    .all()
  return render_template('wiki/all_pages.html', pages=pages)


@route('/help')
def wiki_help():
  filename = join(dirname(__file__), 'data', 'help.txt')
  src = open(filename).read()
  body = Markup(markdown(src))
  return render_template('wiki/help.html', body=body)


@route('/export')
def wiki_export():
  return "Not done yet"
  # TODO


#
# Util
#
def get_page_by_title(title):
  title = title.strip()
  page = WikiPage.query\
      .filter(WikiPage.community_id == g.community.id,
              WikiPage.title == title)\
      .one()
  return page


def create_home_page():
  default_src = open(join(dirname(__file__), 'data', 'default_page.txt'))\
    .read()\
    .decode('utf-8')
  page = WikiPage(title=u"Home", body_src=default_src)
  page.community_id = g.community.id
  db.session.add(page)
  db.session.commit()
  return page



import re

from flask import flash, current_app, request, g
from flask_babel import gettext as _
from werkzeug.exceptions import NotFound, Forbidden, InternalServerError
from abilian.core.signals import activity
from abilian.services.security import security, Admin, WRITE, MANAGE
from abilian.web import url_for

from ..repository import repository

#
# Utils
#
def breadcrumbs_for(object):
  if object is None:
    return []

  bc = [dict(label=object.title)]
  parent = object.parent
  while parent and not parent.is_root_folder:
    bc = [dict(label=parent.title, path=url_for(parent))] + bc
    parent = parent.parent

  return bc


def get_document(id):
  """
  Gets a document given its id. Will raise appropriates errors in case
  the document doesn't exist (404), or the current user doesn't have read access
  on the document (403).
  """
  doc = repository.get_document_by_id(id)
  check_read_access(doc)
  return doc


def get_folder(id):
  """
  Gets a folder given its id. Will raise appropriates errors in case
  the folder doesn't exist (404), or the current user doesn't have read access
  on the folder (403).
  """
  folder = repository.get_folder_by_id(id)
  check_read_access(folder)
  return folder


def get_new_filename(folder, name):
  """
  Given a desired name for a new content in folder, return a name suitable
  for new content.

  If name already exists, a numbered suffix is added.
  """
  existing = set((c.name for c in folder.children))
  renamed = name in existing

  if renamed:
    name = name.rsplit('.', 1)
    ext = u'.{}'.format(name[1]) if len(name) > 1 else u''
    name = name[0]
    prefix = u'{}-'.format(name)
    prefix_len = len(prefix)
    # find all numbered suffixes from name-1.ext, name-5.ext,...
    suffixes = (n[prefix_len:].rsplit(u'.', 1)[0]
                for n in existing
                if n.startswith(prefix) and n.endswith(ext))
    suffixes = [int(val) for val in suffixes
                if re.match(r'^\d+$', val)]

    index = max(0, 0, *suffixes) + 1 #  0, 0: in case suffixes is empty
    name = u'{}-{}{}'.format(name, index, ext)

  return name


def create_document(folder, fs):
  check_write_access(folder)

  if isinstance(fs.filename, unicode):
    name = fs.filename
  else:
    name = unicode(fs.filename, errors='ignore')

  if not name:
    flash(_(u"Document name can't be empty."), "error")
    return None

  original_name = name
  name = get_new_filename(folder, name)
  doc = folder.create_document(title=name)
  doc.set_content(fs.read(), fs.content_type)

  if original_name != name:
    # set message after document has been successfully created!
    flash(_(u'"{original}" already present in folder, '
            'renamed "{name}"').format(original=original_name, name=name),
          'info')

  # Some unwrapping before posting event
  app = current_app._get_current_object()
  community = g.community._model
  activity.send(app, actor=g.user, verb="post", object=doc, target=community)

  return doc


def edit_object(obj):
  title = request.form.get("title", "")
  description = request.form.get("description", "")

  changed = False
  if title != obj.title:
    obj.title = title
    changed = True
  if description != obj.description:
    obj.description = description
    changed = True

  return changed


def get_selected_objects(folder):
  """
  Returns a tuple, (folders, docs), of folders and docs in the specified
  folder that have been selected from the UI.
  """
  selected_ids = request.form.getlist("object-selected")

  doc_ids = [int(x.split(":")[-1]) for x in selected_ids
             if x.startswith("cmis:document")]
  folder_ids = [int(x.split(":")[-1]) for x in selected_ids
                if x.startswith("cmis:folder")]

  docs = map(get_document, doc_ids)
  folders = map(get_folder, folder_ids)

  for obj in docs + folders:
    if obj.parent != folder:
      raise InternalServerError()

  return folders, docs


def check_read_access(obj):
  """
  Checks the current user has appropriate read access on the given object.
  Will raise appropriates errors in case the object doesn't exist (404),
  or the current user doesn't have read access on the object (403).
  """
  if not obj:
    raise NotFound()
  if not security.running:
    return True
  if security.has_role(g.user, Admin):
    return True
  if repository.has_access(g.user, obj):
    return True
  raise Forbidden()


def check_write_access(obj):
  """
  Checks the current user has appropriate write access on the given object.
  Will raise appropriates errors in case the object doesn't exist (404),
  or the current user doesn't have write access on the object (403).
  """
  if not obj:
    raise NotFound()
  if not security.running:
    return
  if security.has_role(g.user, Admin):
    return

  if (repository.has_access(g.user, obj)
      and repository.has_permission(g.user, WRITE, obj)):
    return
  raise Forbidden()


def check_manage_access(obj):
  """
  Checks the current user has appropriate manage access on the given object.
  Will raise appropriates errors in case the object doesn't exist (404),
  or the current user doesn't have manage access on the object (403).
  """

  if not obj:
    raise NotFound()
  if not security.running:
    return
  if security.has_role(g.user, Admin):
    return
  if (repository.has_access(g.user, obj)
      and repository.has_permission(g.user, MANAGE, obj)):
    return
  raise Forbidden()


def match(mime_type, patterns):
  if not mime_type:
    mime_type = "application/binary"
  for pat in patterns:
    pat = pat.replace("*", r"\w*")
    if re.match(pat, mime_type):
      return True
  return False



# coding=utf-8
"""
"""
from __future__ import absolute_import


from abilian.services.security import security, Reader, Writer, Manager
from abilian.sbe.apps.communities.models import VALID_ROLES
from abilian.sbe.apps.communities.signals import (
  membership_set, membership_removed
)

from .search import reindex_tree

@membership_set.connect
def new_community_member(community, membership, is_new, **kwargs):
  if not community.folder:
    return

  role = membership.role
  user = membership.user
  local_role = Writer if community.type == 'participative' else Reader
  if role == Manager:
    local_role = Manager

  current_roles = set(security.get_roles(user, community.folder,
                                         no_group_roles=True))
  current_roles &= VALID_ROLES  # ensure we don't remove roles not managed
                                # by us
                                
  for role_to_ungrant in current_roles - set((local_role,)):
    security.ungrant_role(user, role_to_ungrant, community.folder)

  if local_role not in current_roles:
    security.grant_role(user, local_role, community.folder)

  reindex_tree(community.folder)
    
@membership_removed.connect
def remove_community_member(community, membership, **kwargs):
  if not community.folder:
    return

  user = membership.user
  roles = set(security.get_roles(user, community.folder, no_group_roles=True))
  roles &= VALID_ROLES  # ensure we don't remove roles not managed by us
  for role in roles:
    security.ungrant_role(user, role, community.folder)
  
  reindex_tree(community.folder)



# coding=utf-8
"""

"""
from __future__ import absolute_import

import sys
import logging
import fileinput
from email.parser import FeedParser

from flask_script import Manager

from .tasks import process_email, check_maildir

logger = logging.getLogger(__name__)


manager = Manager(description='SBE forum commands',
                  help='SBE forum commands')


@manager.option('-f', '--filename',
                help='email filename; defaults to standard input',
                default=u'-', required=False,)
def inject_email(filename=u'-'):
  """
  Reads one email from stdin,
  parse it,
  forward it in a celery task to be persisted.
  """
  parser = FeedParser()
  message = None

  if logger.level is logging.NOTSET:
    logger.setLevel(logging.INFO)

  try:
    # iterate over stdin
    for line in fileinput.input(filename):
      parser.feed(line)
  except KeyboardInterrupt:
    logger.info('Aborted by user, exiting.')
    sys.exit(1)
  except:
    logger.error('Error during email parsing', exc_info=True)
    sys.exit(1)
  finally:
    # close the parser to generate a email.message
    message = parser.close()
    fileinput.close()

  if message:
    # make sure no email.errors are present
    if len(message.defects) == 0:
      process_email.delay(message)
    else:
      logger.error('email has defects, message content:\n'
                   '------ START -------\n'
                   '%s'
                   '\n------ END -------\n',
                   message,
                   extra={ 'stack': True, })
  else:
    logger.error('no email was parsed from stdin', extra={ 'stack': True, })


@manager.command
def check_email():
  """
  Reads one email from current user Maildir,
  parse it,
  forward it in a celery task to be persisted.
  """

  check_maildir.delay()



# coding=utf-8
"""
"""
from __future__ import absolute_import

from datetime import timedelta, datetime

from flask import current_app
from flask_mail import Message
from flask_security.utils import md5
from sqlalchemy import and_, or_
from validate_email import validate_email
from celery import shared_task
from celery.schedules import crontab

from abilian.i18n import render_template_i18n
from abilian.core.models.subjects import User
from abilian.web import url_for
from abilian.services.activity import ActivityEntry
from abilian.services.auth.views import get_serializer

from abilian.sbe.apps.documents.models import Document
from abilian.sbe.apps.documents.repository import repository
from abilian.sbe.apps.forum.models import Thread, Post
from abilian.sbe.apps.wiki.models import WikiPage
from .. import TOKEN_SERIALIZER_NAME


DIGEST_TASK_NAME = __name__ + '.send_daily_social_digest_task'
DEFAULT_DIGEST_SCHEDULE = {
  'task': DIGEST_TASK_NAME,
  'schedule': crontab(hour=10, minute=0, )
}


# expires after 1 day - 10 minutes
@shared_task(expires=85800)
def send_daily_social_digest_task():
  # a request_context is required when rendering templates
  with current_app.test_request_context('/send_daily_social_updates'):
    config = current_app.config
    if not config.get('PRODUCTION') or config.get("DEMO"):
      return
    send_daily_social_digest()


def send_daily_social_digest():
  for user in User.query.filter(User.can_login == True).all():
    preferences = current_app.services['preferences']
    prefs = preferences.get_preferences(user)

    if not prefs.get('sbe:notifications:daily', False):
      continue

    # Defensive programming.
    if not validate_email(user.email):
      continue

    try:
      send_daily_social_digest_to(user)
    except:
      current_app.logger.error('Error sending daily social digest', exc_info=True)


def send_daily_social_digest_to(user):
  """Send to a given user a daily digest of activities in its communities.

  Return 1 if mail sent, 0 otherwise.
  """
  mail = current_app.extensions['mail']

  message = make_message(user)
  if message:
    mail.send(message)
    return 1
  else:
    return 0


def make_message(user):
  config = current_app.config
  sbe_config = config['ABILIAN_SBE']
  sender = config.get('BULK_MAIL_SENDER', config['MAIL_SENDER'])
  recipient = user.email
  subject = sbe_config['DAILY_SOCIAL_DIGEST_SUBJECT']
  digests = []
  happened_after = datetime.utcnow() - timedelta(days=1)
  list_id = u'"{} daily digest" <daily.digest.{}>'.format(
    config['SITE_NAME'],
    config.get('SERVER_NAME', u'example.com'),
  )
  base_extra_headers = {
    'List-Id': list_id,
    'List-Post': 'NO',
    'Auto-Submitted': 'auto-generated',
    'X-Auto-Response-Suppress': 'All',
    'Precedence': 'bulk',
  }

  for membership in user.communautes_membership:
    community = membership.community
    if not community:
      # TODO: should not happen but it does. Fix root cause instead.
      continue
    # create an empty digest
    digest = CommunityDigest(community)
    AE = ActivityEntry
    activities = AE.query\
      .order_by(AE.happened_at.asc()) \
      .filter(
        and_(AE.happened_at > happened_after,
             or_(and_(AE.target_type == community.object_type,
                      AE.target_id == community.id),
                 and_(AE.object_type == community.object_type,
                      AE.object_id == community.id),))
      ) \
      .all()

    # fill the internal digest lists with infos
    # seen_entities, new_members, new_documents, updated_documents ...
    for activity in activities:
      digest.update_from_activity(activity, user)
    # if activities:
    #   import ipdb; ipdb.set_trace()
    # save the current digest in the master digests list
    if not digest.is_empty():
      digests.append(digest)

  if not digests:
    return None

  token = generate_unsubscribe_token(user)
  unsubscribe_url = url_for('notifications.unsubscribe_sbe',
                            token=token,
                            _external=True,
                            _scheme=config['PREFERRED_URL_SCHEME'])
  extra_headers = dict(base_extra_headers)
  extra_headers['List-Unsubscribe'] = u'<{}>'.format(unsubscribe_url)

  msg = Message(subject, sender=sender, recipients=[recipient],
                extra_headers=extra_headers)
  ctx = {'digests': digests, 'token': token, 'unsubscribe_url': unsubscribe_url}
  msg.body = render_template_i18n("notifications/daily-social-digest.txt",
                                  **ctx)
  msg.html = render_template_i18n("notifications/daily-social-digest.html",
                                  **ctx)
  return msg


def generate_unsubscribe_token(user):
  """Generates a unique unsubscription token for the specified user.

  :param user: The user to work with
  """
  data = [str(user.id), md5(user.password)]
  return get_serializer(TOKEN_SERIALIZER_NAME).dumps(data)


class CommunityDigest(object):
  def __init__(self, community):
    self.community = community

    self.seen_entities = set()
    self.new_members = []
    self.new_documents = []
    self.updated_documents = []
    self.new_conversations = []
    self.updated_conversations = {}
    self.new_wiki_pages = []
    self.updated_wiki_pages = {}

  def is_empty(self):
    return (not self.new_members and not self.new_documents and not
            self.updated_documents and not self.new_conversations and not
            self.updated_conversations and not self.new_wiki_pages and not
            self.updated_wiki_pages)

  def update_from_activity(self, activity, user):
    actor = activity.actor
    obj = activity.object

    # TODO ?
    #target = activity.target

    if activity.verb == 'join':
      self.new_members.append(actor)

    elif activity.verb == 'post':
      if obj is None:
        return
      if obj.id in self.seen_entities:
        return
      self.seen_entities.add(obj.id)

      if isinstance(obj, Document) and repository.has_access(user, obj):
        self.new_documents.append(obj)
      elif isinstance(obj, WikiPage):
        self.new_wiki_pages.append(obj)
      elif isinstance(obj, Thread):
        self.new_conversations.append(obj)
      elif isinstance(obj, Post):
        if obj.thread.id not in self.seen_entities:
          # save actor and oldest/first modified Post in thread
          # oldest post because Activities are ordered_by Asc(A.happened_at)
          self.updated_conversations[obj.thread] = {'actors': [actor],
                                                    'post': obj}
          # Mark this post's Thread as seen to avoid duplicates
          self.seen_entities.add(obj.thread.id)
        elif obj.thread not in self.new_conversations:
          # this post's Thread has already been seen in another Activity
          # exclude it to avoid duplicates but save the Post's actor
          self.updated_conversations[obj.thread]['actors'].append(actor)

    elif activity.verb == 'update':
      if obj is None:
        return
      # special case for Wikipage, we want to know each updater
      if isinstance(obj, WikiPage):
        if obj in self.updated_wiki_pages:
          page = self.updated_wiki_pages[obj]
          if actor in page:
            page[actor] += 1
          else:
            page[actor] = 1
        else:
          self.updated_wiki_pages[obj] = {actor: 1}

      # fast return for all other objects
      if obj.id in self.seen_entities:
        return
      self.seen_entities.add(obj.id)

      # all objects here need to be accounted only once
      if isinstance(obj, Document) and repository.has_access(user, obj):
        self.updated_documents.append(obj)

